==> sslexplorer-0.2.13-src.tar.gz: sslexplorer-0.2.13/sslexplorer/src/com/sslexplorer/core/RequestParameterMap.java <==
/*
 *  SSL-Explorer
 *
 *  Copyright (C) 2003-2006 3SP LTD. All Rights Reserved
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
			

/*
 * Parts of the code that deal with multipart/form-data parsing are (loosely)
 * based on Jettys MultipartRequest. See license below.
 *
 * 24/02/2005 - brett@3sp.com
 *
 */

//========================================================================
//$Id$
//Copyright 1996-2004 Mort Bay Consulting Pty. Ltd.
//------------------------------------------------------------------------
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//http://www.apache.org/licenses/LICENSE-2.0
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.
//========================================================================
package com.sslexplorer.core;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.sslexplorer.boot.RequestHandlerRequest;
import com.sslexplorer.boot.Util;
import com.sslexplorer.policyframework.LaunchSession;

/**
 * 
 * Extracts request parameters given a {@link RequestHandlerRequest}.
 * <p>
 * If the request has a <i>Content type</i> of <b>multipart/form-data</b>,
 * then parameters are read from the input stream. Requests of this type will
 * return <code>true</code> for {@link #isFormData()}.
 * <p>
 * If the request does not have a content type of <b>multipart/form-data</b>,
 * then the request parameters are read directly from the
 * {@link RequestHandlerRequest#getParameters()} method. Requests of this type
 * will return <code>false</code> for {@link #isFormData()}.
 * <p>
 * <b>IMPORTANT</b>. Names and values stored in this map must <b>NOT</b> be
 * encoded in any way. Names and values should be encoded as and when needed.
 * <p>
 * The <i>Proxied URL</i> (i.e. the actual URL to load) will also be extracted
 * from the request and can be retrieved using {@link #getProxiedURL()}.
 * 
 * @author Lee David Painter <a href="mailto:
 *         brett@3sp.com">&lt;brett@3sp.com&gt;</a>
 * @author Brett Smith <a href="mailto: brett@3sp.com">&lt;brett@3sp.com&gt;</a>
 */
public class RequestParameterMap extends HashMap<String, String> {

	final static Log log = LogFactory.getLog(RequestParameterMap.class);

	/**
	 * Get the encoding name for ISO_8859_1. This may be configured using the
	 * system property <code>ISO_8859_1</code>.
	 */
	public final static String ISO_8859_1;
	static {
		String iso = System.getProperty("ISO_8859_1");
		if (iso != null)
			ISO_8859_1 = iso;
		else {
			try {
				new String(new byte[] { (byte) 20 }, "ISO-8859-1");
				iso = "ISO-8859-1";
			} catch (java.io.UnsupportedEncodingException e) {
				iso = "ISO8859_1";
			}
			ISO_8859_1 = iso;
		}
	}

	private static final long serialVersionUID = -2005125971106236972L;

	// Private instance variables

	private LineInput input;
	private String boundary;
	private byte[] boundaryBytes;
	private boolean finished;
	private File multipartFile;
	private long multipartLength;
	private int ch;
	private URL proxiedURLBase, proxiedURL;
	private String launchId;

	/**
	 * Constructor.
	 */
	public RequestParameterMap() {
		super();
		ch = -2;
	}

	/**
	 * Constructor.
	 * 
	 * @param request request to extract parameters from
	 * @throws IOException
	 */
	public RequestParameterMap(RequestHandlerRequest request) throws IOException {
		this();
		String contentType = request.getField("content-type");
		
		if (contentType != null && contentType.startsWith("multipart/form-data")) {
			initMultipart(contentType, request.getInputStream());
		} else {
			// Parameter names and values from the request are not encoded so
			// can be placed straight in the map
			Map params = request.getParameters();
			for (Iterator e = params.keySet().iterator(); e.hasNext();) {
				String n = (String) e.next();

				Object obj = params.get(n);

				if (!checkForSSLXUrl(n, obj)) {
					if (obj instanceof String[]) {
						put(n, ((String[]) obj)[0]);
					} else if (obj instanceof String) {
						put(n, obj.toString());
					} else
						log.warn("Parameter value is an unexepected type " + obj.getClass().getName());
				}
			}
		}
	}

	/**
	 * Get the ID of the {@link LaunchSession} that launched the web forward
	 * this request is part of. This is specified by the request parameter with
	 * the name of the contstant {@link LaunchSession#LONG_LAUNCH_ID} value.
	 * 
	 * @return launch ID
	 */
	public String getLaunchId() {
		return launchId;
	}

	/**
	 * Get if this map was initialised from a request with a content type of
	 * <b>multipart/form-data</b>.
	 * 
	 * @return is form data
	 */
	public boolean isFormData() {
		return multipartFile != null;
	}

	/**
	 * If this map was initialised from a <b>multipart/form-data</b> request,
	 * then this method returns an input stream of the content received. This
	 * content is actually read from a temporary file which is deleted when the
	 * stream is closed.
	 * 
	 * @return form data stream
	 * @throws IOException if map not initialised from <b>multipart/form-data</b>
	 *         request
	 * @throws FileNotFoundException
	 */
	public InputStream getFormData() throws IOException, FileNotFoundException {
		if (multipartFile == null) {
			throw new IOException("This request was not of type multipart/form-data.");
		}
		return new FileMultipartInputStream(multipartFile);
	}

	/**
	 * If this map was initialised from a <b>multipart/form-data</b> request,
	 * then this method returns the number of bytes in the content.
	 * 
	 * @return form data length
	 * @throws IOException
	 */
	public long getFormDataLength() throws IOException {
		if (multipartFile == null) {
			throw new IOException("This request was not of type multipart/form-data.");
		}
		return multipartLength;
	}

	/**
	 * Get the <i>Proxied URL</i>, i.e. the URL of the page to actually load
	 * from the target server. This will include any request parameters supplied
	 * when a GET rquest.
	 * 
	 * @return proxied URL
	 */
	public URL getProxiedURL() {
		return proxiedURL;
	}

	/**
	 * Get the <i>Proxied URL base</i>, i.e. the URL of the page to actually
	 * load from the target server <b>excluding</b> any request parameters if
	 * the request is a GET.
	 * 
	 * @return proxied URL base
	 */
	public URL getProxiedURLBase() {
		return proxiedURLBase;
	}

	/**
	 * Get a parameters value given its name. The value returned is unencoded.
	 * 
	 * @param name parameter name
	 * @return encoded value
	 */
	public String getParameter(Object name) {
		return get(name);
	}

	/**
	 * Get an {@link Iterator} of all parameters names contained in this map.
	 * 
	 * @return parameters name
	 */
	public Iterator getParameterNames() {
		return keySet().iterator();
	}

	/*
	 * public String[] getParameterValues(String name) { return (String[])
	 * super.get(name); }
	 */

	private void readBytes(InputStream in, OutputStream out, OutputStream recorded) throws IOException {

		int c;
		boolean cr = false;
		boolean lf = false;

		// loop for all lines`
		while (true) {
			int b = 0;
			while ((c = (ch != -2) ? ch : readWrite(in, recorded)) != -1) {
				multipartLength++;
				ch = -2;

				// look for CR and/or LF
				if (c == 13 || c == 10) {
					if (c == 13) {
						ch = readWrite(in, recorded);
						multipartLength++;
					}
					break;
				}

				// look for boundary
				if (b >= 0 && b < boundaryBytes.length && c == boundaryBytes[b])
					b++;
				else {
					// this is not a boundary
					if (cr && out != null)
						out.write(13);
					if (lf && out != null)
						out.write(10);
					cr = lf = false;

					if (b > 0 && out != null)
						out.write(boundaryBytes, 0, b);
					b = -1;

					if (out != null)
						out.write(c);
				}
			}

			// check partial boundary
			if ((b > 0 && b < boundaryBytes.length - 2) || (b == boundaryBytes.length - 1)) {
				if (cr && out != null)
					out.write(13);
				if (lf && out != null)
					out.write(10);
				cr = lf = false;
				if (out != null)
					out.write(boundaryBytes, 0, b);
				b = -1;
			}

			// boundary match
			if (b > 0 || c == -1) {
				if (b == boundaryBytes.length)
					finished = true;
				if (ch == 10)
					ch = -2;
				break;
			}

			// handle CR LF
			if (cr && out != null)
				out.write(13);
			if (lf && out != null)
				out.write(10);
			cr = (c == 13);
			lf = (c == 10 || ch == 10);
			if (ch == 10)
				ch = -2;
		}
	}

	private int readWrite(InputStream in, OutputStream out) throws IOException {
		int i = in.read();
		if (i != -1 && out != null) {
			out.write(i);
		}
		return i;

	}

	private boolean checkForSSLXUrl(String n, Object params) {
		String val;

		if (params instanceof String[]) {
			val = ((String[]) params)[0];
		} else
			val = (String) params;

		if (n.equals("sslex_url")) {
			parseProxiedURL(val);
			return true;
		} else if (n.equals(LaunchSession.LONG_LAUNCH_ID)) {
			launchId = val;
			return true;
		}
		return false;
	}

	private void initMultipart(String contentType, InputStream in) throws IOException, FileNotFoundException {
		multipartFile = File.createTempFile("mpr", ".tmp");
		FileOutputStream mpOut = new FileOutputStream(multipartFile);

		try {

			input = new LineInput(in);
			boundary = "--" + Util.valueOfNameValuePair(contentType.substring(contentType.indexOf("boundary=")));
			boundaryBytes = (boundary + "--").getBytes(ISO_8859_1);// Get
			// first
			// boundary
			String line = input.readLine();
			if (!line.equals(boundary)) {
				throw new IOException("Missing initial multi part boundary");
			}

			byte[] buf = (line + "\r\n").getBytes();
			multipartLength += buf.length;
			mpOut.write(buf);

			while (!finished) {
				String contentDisposition = null;

				Map<String, List<String>> parms = new TreeMap<String, List<String>>();
				while ((line = input.readLine()) != null) {
					if (line.length() == 0)
						break;
					int idx = line.indexOf(':', 0);
					if (idx > 0) {
						String key = line.substring(0, idx).trim();
						String value = line.substring(idx + 1, line.length()).trim();

						List<String> cur = (List<String>) parms.get(key);
						if (cur == null) {
							cur = new ArrayList<String>();
							parms.put(key, cur);
						}
						cur.add(value);

						if (key.equalsIgnoreCase("content-disposition"))
							contentDisposition = value;
					}
				}
				boolean form = false;
				if (contentDisposition == null) {
					throw new IOException("Missing content-disposition");
				}
				StringTokenizer tok = new StringTokenizer(contentDisposition, ";");
				String name = null;
				String filename = null;
				while (tok.hasMoreTokens()) {
					String t = tok.nextToken().trim();
					String tl = t.toLowerCase();
					if (t.startsWith("form-data"))
						form = true;
					else if (tl.startsWith("name="))
						name = Util.valueOfNameValuePair(t);
					else if (tl.startsWith("filename="))
						filename = Util.valueOfNameValuePair(t);
				}

				// Check disposition
				if (!form) {
					log.warn("Non form-data part in multipart/form-data");
					continue;
				}

				if (name.equals("sslex_url")) {
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					readBytes(input, baos, null);
					parseProxiedURL(new String(baos.toString()));
				} else {
					for (Iterator i = parms.keySet().iterator(); i.hasNext();) {
						String key = (String) i.next();
						List list = (List) parms.get(key);
						for (Iterator j = list.iterator(); j.hasNext();) {
							String val = (String) j.next();
							buf = (key + ": " + val + "\r\n").getBytes();
							multipartLength += buf.length;
							mpOut.write(buf);
						}
					}
					buf = "\r\n".getBytes();
					multipartLength += buf.length;
					mpOut.write(buf);

					if (filename != null) {
						readBytes(input, null, mpOut);
					} else {
						ByteArrayOutputStream baos = new ByteArrayOutputStream();
						readBytes(input, baos, mpOut);
						put(Util.urlDecode(name), baos.toString());
					}
				}
			}
		} finally {
			Util.closeStream(mpOut);
		}

	}
	
	public void parseProxiedURL(String location) {
		try {
			proxiedURL = new URL(location);

			// Extract parameters from the proxied URL
			String query = proxiedURL.getQuery();
			if (query != null) {
				proxiedURLBase = new URL(proxiedURL.getProtocol(), proxiedURL.getHost(), proxiedURL.getPort() < 1 ? -1 : proxiedURL.getPort() , proxiedURL.getFile());
				StringTokenizer t = new StringTokenizer(query, "&");
				while (t.hasMoreTokens()) {
					String parm = t.nextToken();
					int pidx = parm.indexOf('=');
					String name = pidx == -1 ? parm : parm.substring(0, pidx);
					String value = pidx == -1 ? "" : parm.substring(pidx + 1);
					put(name, value);
				}
			} else {
				proxiedURLBase = proxiedURL;
			}
		} catch (MalformedURLException murle) {
			log.error("Invalid proxied URL '" + location + "'");
		}
	}

	class FileMultipartInputStream extends FileInputStream {

		private File file;

		FileMultipartInputStream(File file) throws FileNotFoundException {
			super(file);
			this.file = file;
		}

		public void close() throws IOException {
			super.close();
			file.delete();
		}

	}
}

==> sslexplorer-0.2.15_01-src.tar.gz: sslexplorer-0.2.15_01/sslexplorer/src/com/sslexplorer/core/RequestParameterMap.java <==
/*
 *  SSL-Explorer
 *
 *  Copyright (C) 2003-2006 3SP LTD. All Rights Reserved
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
			

/*
 * Parts of the code that deal with multipart/form-data parsing are (loosely)
 * based on Jettys MultipartRequest. See license below.
 *
 * 24/02/2005 - brett@3sp.com
 *
 */

//========================================================================
//$Id$
//Copyright 1996-2004 Mort Bay Consulting Pty. Ltd.
//------------------------------------------------------------------------
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//http://www.apache.org/licenses/LICENSE-2.0
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.
//========================================================================
package com.sslexplorer.core;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mortbay.util.MultiMap;

import com.sslexplorer.boot.RequestHandlerRequest;
import com.sslexplorer.boot.Util;
import com.sslexplorer.policyframework.LaunchSession;

/**
 * 
 * Extracts request parameters given a {@link RequestHandlerRequest}.
 * <p>
 * If the request has a <i>Content type</i> of <b>multipart/form-data</b>,
 * then parameters are read from the input stream. Requests of this type will
 * return <code>true</code> for {@link #isFormData()}.
 * <p>
 * If the request does not have a content type of <b>multipart/form-data</b>,
 * then the request parameters are read directly from the
 * {@link RequestHandlerRequest#getParameters()} method. Requests of this type
 * will return <code>false</code> for {@link #isFormData()}.
 * <p>
 * <b>IMPORTANT</b>. Names and values stored in this map must <b>NOT</b> be
 * encoded in any way. Names and values should be encoded as and when needed.
 * <p>
 * The <i>Proxied URL</i> (i.e. the actual URL to load) will also be extracted
 * from the request and can be retrieved using {@link #getProxiedURL()}.
 * 
 * @author Lee David Painter <a href="mailto:
 *         brett@3sp.com">&lt;brett@3sp.com&gt;</a>
 * @author Brett Smith <a href="mailto: brett@3sp.com">&lt;brett@3sp.com&gt;</a>
 */
public class RequestParameterMap extends MultiMap {

	final static Log log = LogFactory.getLog(RequestParameterMap.class);

	/**
	 * Get the encoding name for ISO_8859_1. This may be configured using the
	 * system property <code>ISO_8859_1</code>.
	 */
	public final static String ISO_8859_1;
	static {
		String iso = System.getProperty("ISO_8859_1");
		if (iso != null)
			ISO_8859_1 = iso;
		else {
			try {
				new String(new byte[] { (byte) 20 }, "ISO-8859-1");
				iso = "ISO-8859-1";
			} catch (java.io.UnsupportedEncodingException e) {
				iso = "ISO8859_1";
			}
			ISO_8859_1 = iso;
		}
	}

	private static final long serialVersionUID = -2005125971106236972L;

	// Private instance variables

	private LineInput input;
	private String boundary;
	private byte[] boundaryBytes;
	private boolean finished;
	private File multipartFile;
	private long multipartLength;
	private int ch;
	private URL proxiedURLBase, proxiedURL;
	private String launchId;

	/**
	 * Constructor.
	 */
	public RequestParameterMap() {
		super();
		ch = -2;
	}

	/**
	 * Constructor.
	 * 
	 * @param request request to extract parameters from
	 * @throws IOException
	 */
	public RequestParameterMap(RequestHandlerRequest request) throws IOException {
		this();
		String contentType = request.getField("content-type");
		
		if (contentType != null && contentType.startsWith("multipart/form-data")) {
			initMultipart(contentType, request.getInputStream());
		}
		
		String requestPath = request.getPath();


			/* The launch session and URL may be provided in one of two ways.
			 * 
			 * 1. As a request to /replacementProxyEngine with the launch session
			 *    and target URL provided as sslx_launchId and sslx_url respectively
			 *    
			 * 2. In the new format /replacementProxyEngine/[launchId]/[encodedURL]
			 */
			
			if(requestPath != null && requestPath.startsWith("/replacementProxyEngine/")) {
				int idx = requestPath.indexOf('/', 1);
				int idx2 = requestPath.indexOf('/', idx + 1);
				launchId = requestPath.substring(idx + 1, idx2);
				parseProxiedURL(requestPath.substring(idx2 + 1));
			}
			
			// Parameter names and values from the request are not encoded so
			// can be placed straight in the map
			Map params = request.getParameters();
			for (Iterator e = params.keySet().iterator(); e.hasNext();) {
				String n = (String) e.next();

				Object obj = params.get(n);

				if(!checkForSSLXUrl(n, obj)) {				
					if (obj instanceof String[]) {
						put(n, ((String[]) obj)[0]);
					} else if (obj instanceof String) {
						put(n, obj.toString());
					} else
						log.warn("Parameter value is an unexepected type " + obj.getClass().getName());
				}
			}
		
	}

	/**
	 * Get the ID of the {@link LaunchSession} that launched the web forward
	 * this request is part of. This is specified by the request parameter with
	 * the name of the contstant {@link LaunchSession#LONG_LAUNCH_ID} value.
	 * 
	 * @return launch ID
	 */
	public String getLaunchId() {
		return launchId;
	}

	/**
	 * Get if this map was initialised from a request with a content type of
	 * <b>multipart/form-data</b>.
	 * 
	 * @return is form data
	 */
	public boolean isFormData() {
		return multipartFile != null;
	}
	
	/**
	 * TODO: This current calls getParameter to ensure that only a single paramter is
	 * returned. Implementations that use this type of Map should be changed to 
	 * support multiple values for a single key, or to use getParameter if they 
	 * do not require multiple value support. 
	 * @param name
	 * @return the first value in the list
	 */
	public Object get(Object name) {
		return getParameter(name);
	}

	/**
	 * If this map was initialised from a <b>multipart/form-data</b> request,
	 * then this method returns an input stream of the content received. This
	 * content is actually read from a temporary file which is deleted when the
	 * stream is closed.
	 * 
	 * @return form data stream
	 * @throws IOException if map not initialised from <b>multipart/form-data</b>
	 *         request
	 * @throws FileNotFoundException
	 */
	public InputStream getFormData() throws IOException, FileNotFoundException {
		if (multipartFile == null) {
			throw new IOException("This request was not of type multipart/form-data.");
		}
		return new FileMultipartInputStream(multipartFile);
	}

	/**
	 * If this map was initialised from a <b>multipart/form-data</b> request,
	 * then this method returns the number of bytes in the content.
	 * 
	 * @return form data length
	 * @throws IOException
	 */
	public long getFormDataLength() throws IOException {
		if (multipartFile == null) {
			throw new IOException("This request was not of type multipart/form-data.");
		}
		return multipartLength;
	}

	/**
	 * Get the <i>Proxied URL</i>, i.e. the URL of the page to actually load
	 * from the target server. This will include any request parameters supplied
	 * when a GET rquest.
	 * 
	 * @return proxied URL
	 */
	public URL getProxiedURL() {
		return proxiedURL;
	}

	/**
	 * Get the <i>Proxied URL base</i>, i.e. the URL of the page to actually
	 * load from the target server <b>excluding</b> any request parameters if
	 * the request is a GET.
	 * 
	 * @return proxied URL base
	 */
	public URL getProxiedURLBase() {
		return proxiedURLBase;
	}

	/**
	 * Get a parameters value given its name. The value returned is unencoded.
	 * 
	 * @param name parameter name
	 * @return encoded value
	 */
	public String getParameter(Object name) {
		List list = getValues(name);
		if(list!=null && list.size() >= 1) {
			return (String) list.get(0);
		} else
			return null;
	}
	
	public List getParameterValues(Object name) {
		return getValues(name);
	}

	/**
	 * Get an {@link Iterator} of all parameters names contained in this map.
	 * 
	 * @return parameters name
	 */
	public Iterator getParameterNames() {
		return keySet().iterator();
	}

	/*
	 * public String[] getParameterValues(String name) { return (String[])
	 * super.get(name); }
	 */

	private void readBytes(InputStream in, OutputStream out, OutputStream recorded) throws IOException {

		int c;
		boolean cr = false;
		boolean lf = false;

		// loop for all lines`
		while (true) {
			int b = 0;
			while ((c = (ch != -2) ? ch : readWrite(in, recorded)) != -1) {
				multipartLength++;
				ch = -2;

				// look for CR and/or LF
				if (c == 13 || c == 10) {
					if (c == 13) {
						ch = readWrite(in, recorded);
						multipartLength++;
					}
					break;
				}

				// look for boundary
				if (b >= 0 && b < boundaryBytes.length && c == boundaryBytes[b])
					b++;
				else {
					// this is not a boundary
					if (cr && out != null)
						out.write(13);
					if (lf && out != null)
						out.write(10);
					cr = lf = false;

					if (b > 0 && out != null)
						out.write(boundaryBytes, 0, b);
					b = -1;

					if (out != null)
						out.write(c);
				}
			}

			// check partial boundary
			if ((b > 0 && b < boundaryBytes.length - 2) || (b == boundaryBytes.length - 1)) {
				if (cr && out != null)
					out.write(13);
				if (lf && out != null)
					out.write(10);
				cr = lf = false;
				if (out != null)
					out.write(boundaryBytes, 0, b);
				b = -1;
			}

			// boundary match
			if (b > 0 || c == -1) {
				if (b == boundaryBytes.length)
					finished = true;
				if (ch == 10)
					ch = -2;
				break;
			}

			// handle CR LF
			if (cr && out != null)
				out.write(13);
			if (lf && out != null)
				out.write(10);
			cr = (c == 13);
			lf = (c == 10 || ch == 10);
			if (ch == 10)
				ch = -2;
		}
	}

	private int readWrite(InputStream in, OutputStream out) throws IOException {
		int i = in.read();
		if (i != -1 && out != null) {
			out.write(i);
		}
		return i;

	}

	private boolean checkForSSLXUrl(String n, Object params) {
		String val;

		if (params instanceof String[]) {
			val = ((String[]) params)[0];
		} else
			val = (String) params;

		if (n.equals("sslex_url")) {
			parseProxiedURL(val);
			return true;
		} else if (n.equals(LaunchSession.LONG_LAUNCH_ID)) {
			launchId = val;
			return true;
		}
		return false;
	}

	private void initMultipart(String contentType, InputStream in) throws IOException, FileNotFoundException {
		multipartFile = File.createTempFile("mpr", ".tmp");
		FileOutputStream mpOut = new FileOutputStream(multipartFile);

		try {

			input = new LineInput(in);
			boundary = "--" + Util.valueOfNameValuePair(contentType.substring(contentType.indexOf("boundary=")));
			boundaryBytes = (boundary + "--").getBytes(ISO_8859_1);// Get
			// first
			// boundary
			String line = input.readLine();
			if (!line.equals(boundary)) {
				throw new IOException("Missing initial multi part boundary");
			}

			byte[] buf = (line + "\r\n").getBytes();
			multipartLength += buf.length;
			mpOut.write(buf);

			while (!finished) {
				String contentDisposition = null;

				Map<String, List<String>> parms = new TreeMap<String, List<String>>();
				while ((line = input.readLine()) != null) {
					if (line.length() == 0)
						break;
					int idx = line.indexOf(':', 0);
					if (idx > 0) {
						String key = line.substring(0, idx).trim();
						String value = line.substring(idx + 1, line.length()).trim();

						List<String> cur = (List<String>) parms.get(key);
						if (cur == null) {
							cur = new ArrayList<String>();
							parms.put(key, cur);
						}
						cur.add(value);

						if (key.equalsIgnoreCase("content-disposition"))
							contentDisposition = value;
					}
				}
				boolean form = false;
				if (contentDisposition == null) {
					throw new IOException("Missing content-disposition");
				}
				StringTokenizer tok = new StringTokenizer(contentDisposition, ";");
				String name = null;
				String filename = null;
				while (tok.hasMoreTokens()) {
					String t = tok.nextToken().trim();
					String tl = t.toLowerCase();
					if (t.startsWith("form-data"))
						form = true;
					else if (tl.startsWith("name="))
						name = Util.valueOfNameValuePair(t);
					else if (tl.startsWith("filename="))
						filename = Util.valueOfNameValuePair(t);
				}

				// Check disposition
				if (!form) {
					log.warn("Non form-data part in multipart/form-data");
					continue;
				}

				if (name.equals("sslex_url")) {
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					readBytes(input, baos, null);
					parseProxiedURL(new String(baos.toString()));
				} else {
					for (Iterator i = parms.keySet().iterator(); i.hasNext();) {
						String key = (String) i.next();
						List list = (List) parms.get(key);
						for (Iterator j = list.iterator(); j.hasNext();) {
							String val = (String) j.next();
							buf = (key + ": " + val + "\r\n").getBytes();
							multipartLength += buf.length;
							mpOut.write(buf);
						}
					}
					buf = "\r\n".getBytes();
					multipartLength += buf.length;
					mpOut.write(buf);

					if (filename != null) {
						readBytes(input, null, mpOut);
					} else {
						ByteArrayOutputStream baos = new ByteArrayOutputStream();
						readBytes(input, baos, mpOut);
						put(Util.urlDecode(name), baos.toString());
					}
				}
			}
		} finally {
			Util.closeStream(mpOut);
		}

	}
	
	public void parseProxiedURL(String location) {
		try {
			proxiedURL = new URL(location);

			// Extract parameters from the proxied URL
			String query = proxiedURL.getQuery();
			if (query != null) {
				proxiedURLBase = new URL(proxiedURL.getProtocol(), proxiedURL.getHost(), proxiedURL.getPort() < 1 ? -1 : proxiedURL.getPort() , proxiedURL.getFile());
				StringTokenizer t = new StringTokenizer(query, "&");
				while (t.hasMoreTokens()) {
					String parm = t.nextToken();
					int pidx = parm.indexOf('=');
					String name = pidx == -1 ? parm : parm.substring(0, pidx);
					String value = pidx == -1 ? "" : parm.substring(pidx + 1);
					put(name, value);
				}
			} else {
				proxiedURLBase = proxiedURL;
			}
		} catch (MalformedURLException murle) {
			log.error("Invalid proxied URL '" + location + "'");
		}
	}

	class FileMultipartInputStream extends FileInputStream {

		private File file;

		FileMultipartInputStream(File file) throws FileNotFoundException {
			super(file);
			this.file = file;
		}

		public void close() throws IOException {
			super.close();
			file.delete();
		}

	}
}
